#ifndef __MD_LIB_MSGT_H__
#define __MD_LIB_MSGT_H__

#include <cassert>
#include <string>

#include "MD/MD_Message.h"
#include "MD/MD_DateTime.h"
#include "MD/MD_numeric.h"
#include "MD/MD_Types.h"
#include "MD/MD_Protocols.h"

namespace MD {

class MsgT : public MD_Message
{
  public:
  typedef MD_MsgIterator<MsgT> iterator;
  class Field;

  MsgT(bool opt = true);
  MsgT(bool opt, const MD_Message::Allocator& allocator);
  MsgT(const MsgT& src);
  MsgT(MsgT&& src);
  MsgT(const unsigned char* ptr, uint32_t size);
  MsgT(const unsigned char* ptr,
       uint32_t size,
       const MD_Message::Allocator& allocator);
  virtual ~MsgT();

  MsgT & operator= (const MsgT & src);

  // Give the user a new empty message of the same type as this one.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> create() const;

  // Give the user a new copy of the message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> clone() const;

  // Give the user a new pointer containing this message.
  // The user is responsible for deleting the pointer.
  virtual std::unique_ptr<MD_Message> move();

  //! Accessor methods
  virtual bool getChar     (FieldID id, char &        result) const;
  virtual bool getInt      (FieldID id, int &         result) const;
  virtual bool getInt64    (FieldID id, int64_t &     result) const;
  virtual bool getFloat    (FieldID id, float &       result) const;
  virtual bool getDouble   (FieldID id, double &      result) const;
  virtual bool getString   (FieldID id, String &      result) const;
  virtual bool getBuffer   (FieldID id, Buffer &      result) const;
  virtual bool getDateTime (FieldID id, MD_DateTime & result) const;
  virtual bool getEpochNS  (FieldID id, MD_EpochNS &  result) const;

  virtual bool addChar     (FieldID id, char        data);
  virtual bool addInt      (FieldID id, int         data);
  virtual bool addInt64    (FieldID id, int64_t     data);
  virtual bool addFloat    (FieldID id, float       data);
  virtual bool addDouble   (FieldID id, double      data);
  virtual bool addString   (FieldID id, String      data);
  virtual bool addBuffer   (FieldID id, Buffer      data);
  virtual bool addDateTime (FieldID id, MD_DateTime data);
  virtual bool addEpochNS  (FieldID id, MD_EpochNS  data);

  virtual bool updateChar     (FieldID id, char        data);
  virtual bool updateInt      (FieldID id, int         data);
  virtual bool updateInt64    (FieldID id, int64_t     data);
  virtual bool updateFloat    (FieldID id, float       data);
  virtual bool updateDouble   (FieldID id, double      data);
  virtual bool updateString   (FieldID id, String      data);
  virtual bool updateBuffer   (FieldID id, Buffer      data);
  virtual bool updateDateTime (FieldID id, MD_DateTime data);
  virtual bool updateEpochNS  (FieldID id, MD_EpochNS  data);

  // Pull in base class convenience methods
  using MD_Message::getString;
  using MD_Message::addString;
  using MD_Message::updateString;

  virtual bool erase (FieldID id);
  virtual bool erase (const std::set<FieldID> &ids);
  virtual bool erase (const std::vector<FieldID> &ids);

  virtual void clear();

  virtual bool empty() const;
  virtual bool exists (FieldID id) const;

  virtual const unsigned char * getBinary() const;
  virtual bool setBinary (const unsigned char * ptr, uint32_t size);

  virtual uint32_t getSize() const;

  virtual PROTOCOLS::MsgProtocol protocol() const;

  virtual const unsigned char *
  getDataPtr (FieldID id, TypeID & type, int & size) const;

  iterator begin() const;
  iterator end()   const;

private:
  virtual void getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const;

  template <typename FieldType>
  bool addField (FieldID id, FieldType data);

  template <typename FieldType>
  bool updateField (FieldID id, FieldType data);

  template <typename FieldType>
  bool getHelper(FieldID id, FieldType & data) const;

  static const size_t DATA_START = 4;
};

class MsgT::Field
{
  typedef MD::DATATYPES::TypeID TypeID;
  typedef MD::FIELDS::FieldID   FieldID;

  public:
  Field(char * ptr);

  FieldID        id() const;
  TypeID       type() const;
  uint32_t     size() const;
  char    *data_ptr() const;

  operator char()    const;
  operator int32_t() const;
  operator int64_t() const;
  operator float()   const;
  operator double()  const;
  operator String()  const;
  operator Buffer()  const;
  operator MD_EpochNS() const;

  bool operator==(const MsgT::Field &rhs) const;
  private:
  static uint8_t header_length();

  char *ptr()        const;
  char *next_ptr()   const;
  char *id_ptr()     const;
  char *size_ptr()   const;
  char *type_ptr()   const;
  bool  isValid()    const;
  void  next();
  friend  class MsgT;
  friend  class MD_MsgIterator<MsgT>;
  friend class MD_Message;
  // state:
  char * m_ptr;
};

// Inline function definitions.

inline MsgT::MsgT(bool opt) : MsgT(opt, MD_Message::DEFAULT_ALLOCATOR) {}

inline MsgT::MsgT(bool opt, const MD_Message::Allocator& allocator)
    : MD_Message(allocator)
{
    m_data->clear();
    m_pOffsets->clear();

    m_data->resize(4);
    m_data->begin()[0] = 'F';
    m_data->begin()[1] = 'D';
    m_data->begin()[2] = 'S';
    m_data->begin()[3] = '1';
}

inline MsgT::MsgT(const MsgT& src) : MD_Message()
{
    *this = src;
}

inline MsgT::MsgT(MsgT&& src) : MD_Message()
{
    swap(src);
}

inline MsgT::MsgT(const unsigned char* ptr, uint32_t size)
    : MsgT(ptr, size, MD_Message::DEFAULT_ALLOCATOR)
{}

inline MsgT::MsgT(const unsigned char* ptr,
                  uint32_t size,
                  const MD_Message::Allocator& allocator)
    : MD_Message(allocator)
{
    setBinary(ptr, size);  // setBinary clears m_data and m_pOffsets
}

inline MsgT::~MsgT() { }

inline std::unique_ptr<MD_Message> MsgT::create() const {
  return std::unique_ptr<MD_Message>(new MsgT);
}

inline std::unique_ptr<MD_Message> MsgT::clone() const {
  return std::unique_ptr<MD_Message>(new MsgT(*this));
}

inline std::unique_ptr<MD_Message> MsgT::move() {
    return std::unique_ptr<MD_Message>(new MsgT(std::move(*this)));
}

inline bool MsgT::getChar(FieldID id, char & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getInt (FieldID id, int & result) const {
  return getHelper<int32_t>(id, result);
}

inline bool MsgT::getInt64 (FieldID id, int64_t & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getFloat (FieldID id, float & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getDouble (FieldID id, double & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getString (FieldID id, String & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getBuffer (FieldID id, Buffer & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getDateTime (FieldID id, MD_DateTime & result) const {
  return getHelper(id, result);
}

inline bool MsgT::getEpochNS (FieldID id, MD_EpochNS & result) const {
  return getHelper(id, result);
}

inline bool MsgT::Field::isValid() const { return m_ptr != NULL; }

inline bool MsgT::exists (FieldID id) const
{ return getField<Field>(id).isValid(); }

inline bool MsgT::addChar(FieldID id, char data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addInt (FieldID id, int data) {
  return !exists(id) ? addField<int32_t>(id, data) : false;
}

inline bool MsgT::addInt64 (FieldID id, int64_t data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addFloat (FieldID id, float data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addDouble (FieldID id, double data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addString (FieldID id, String data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addBuffer (FieldID id, Buffer data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addEpochNS (FieldID id, MD_EpochNS data) {
  return !exists(id) ? addField(id, data) : false;
}

inline bool MsgT::addDateTime (FieldID id, MD_DateTime data) {
  return !exists(id) ? addField(id, data) : false;
}


inline bool MsgT::updateChar(FieldID id, char data) {
  return updateField(id, data);
}

inline bool MsgT::updateInt (FieldID id, int data) {
  return updateField<int32_t>(id, data);
}

inline bool MsgT::updateInt64 (FieldID id, int64_t data) {
  return updateField(id, data);
}

inline bool MsgT::updateFloat (FieldID id, float data) {
  return updateField(id, data);
}

inline bool MsgT::updateDouble (FieldID id, double data) {
  return updateField(id, data);
}

inline bool MsgT::updateString (FieldID id, String data) {
  return updateField(id, data);
}

inline bool MsgT::updateBuffer (FieldID id, Buffer data) {
  return updateField(id, data);
}

inline bool MsgT::updateDateTime (FieldID id, MD_DateTime data) {
  return updateField(id, data);
}

inline bool MsgT::updateEpochNS (FieldID id, MD_EpochNS data) {
  return updateField(id, data);
}

inline void MsgT::clear()
{
  m_data->resize(DATA_START);
  m_pOffsets->clear();
}

inline MsgT::iterator MsgT::begin() const
{ return MsgT::iterator(m_data->begin() + DATA_START); }

inline MsgT::iterator MsgT::end() const
{ return MsgT::iterator(m_data->end()); }

inline uint32_t MsgT::getSize() const
{ return (uint32_t)m_data->size(); }

inline bool MsgT::empty() const
{ return getSize() <= DATA_START; }

inline const unsigned char * MsgT::getBinary() const
{ return reinterpret_cast <unsigned char *> (m_data->begin()); }

inline PROTOCOLS::MsgProtocol MsgT::protocol() const
{ return PROTOCOLS::MSGT; }

inline char *MsgT::Field::size_ptr() const { return m_ptr + 4; }

inline uint32_t MsgT::Field::size() const
{ assert(isValid()); return MD::getint32be(size_ptr()); }

inline char *MsgT::Field::ptr()      const { return m_ptr; }
inline char *MsgT::Field::id_ptr()   const { return m_ptr; }
inline char *MsgT::Field::type_ptr() const { return m_ptr + 2; }
inline char *MsgT::Field::data_ptr() const { return m_ptr + 8; }
inline char *MsgT::Field::next_ptr() const { return data_ptr() + size(); }

inline void MsgT::Field::next() { m_ptr = data_ptr() + size(); }

inline MD::FIELDS::FieldID MsgT::Field::id() const
{ assert(isValid()); return MD::getint16be(id_ptr()); }

inline void MsgT::getFieldIDs(std::back_insert_iterator<std::vector<FieldID> > inserter) const {
  char * ptr = m_data->begin() + DATA_START;
  const char * end_ptr = m_data->end();

  while (ptr < end_ptr) {
    Field field(ptr);
    *inserter++ = field.id();
    ptr = field.next_ptr();
  }
}

// TODO: Make this code resilient to malformed input.  It's
// fragile and prone to buffer overflows, at the moment.
inline MsgT::Field::Field (char * ptr)
  : m_ptr(ptr) { }

inline MD::DATATYPES::TypeID MsgT::Field::type() const
{
  assert(isValid());
  // Type is sent as a uint16, but never goes above 255 in a properly
  // formed message.  We can therefore read type as a uint8_t, and
  // avoid worrying about byte-ordering entirely.
  return *reinterpret_cast <uint8_t *> (type_ptr() + 1);
}

inline uint8_t MsgT::Field::header_length() { return 8; }

inline bool MsgT::Field::operator==(const MsgT::Field &rhs) const
{
  return (type() == rhs.type() && size() == rhs.size() &&
          !memcmp(data_ptr(), rhs.data_ptr(), size()));
}


} // end namespace

#endif // __MD_LIB_MSGT_H__
