#ifndef _OWNERSHIP_QUEUE_H_
#define _OWNERSHIP_QUEUE_H_

#include <memory>
#include <queue>
#include <cassert>
#include <cstddef>

namespace MDS {

  // A wrapper around std::queue<T*> that gives safe transfer of
  // ownership semantics without the locking overhead of
  // boost::shared_ptr.  Used properly, this should be much faster than
  // TickMsg, and should be almost as safe.  The same preconditions
  // apply here that apply to std::queue.  (Ex: pop()'ing or front()'ing
  // when empty() is true will probably result in a core dump.)
  template <typename T>
  class OwnershipQueue {
  public:
    typedef T value_type;

    OwnershipQueue() : m_closed(false) { }

    ~OwnershipQueue() { clear(); }

    void clear() {
      while (!empty()) { pop(); }
    }

    [[deprecated("Do not pass raw pointers into OwnershipQueue::push(). Use the other push function which takes a unique_ptr")]]
    void push (T *& pmsg) {
      m_queue.push(pmsg);
      pmsg = NULL; // a little added safety, at least for now.
    }

    // Put a message into the queue.  Transfer control of the pointer
    // from the caller to the queue.
    void push (std::unique_ptr<T> pmsg) {
      m_queue.push(pmsg.release());
    }

    // Return control of a pointer from the queue to the caller, via an
    // unique_ptr.  Once you have the unique_ptr, you may copy it into a shared
    // pointer (which will handle deletion for you), or continue using it as an
    // unique_ptr (which handles deletion for you, but has no copy
    // constructor).
    std::unique_ptr<T> pop() {
      assert(!empty());
      std::unique_ptr<T> pmsg (m_queue.front());
      m_queue.pop();
      return std::move(pmsg);
    }

    // Returns a mutable reference to the element at the front of the
    // queue.
    T & front() {
      assert(!empty());
      return *m_queue.front();
    }

    // Returns a const reference to the element at the front of the
    // queue.
    const T & front() const {
      assert(!empty());
      return *m_queue.front();
    }

    bool is_closed() const { return m_closed; }
    void close() { m_closed = true; }


    // Exactly what you expect from std::queue.
    bool empty() const { return m_queue.empty(); }
    std::size_t size() const { return m_queue.size(); }

    // The queue owns the pointers.  Therefore, there can be no sensible
    // copying of the queue without copying all the data pointed to by
    // all of the pointers.  That would suck, and is almost certainly
    // beyond the scope of this class.  Therefore, don't let the user
    // copy via the copy constructor or by the assignment operator.
    OwnershipQueue(const OwnershipQueue &) = delete;
    OwnershipQueue& operator=(const OwnershipQueue&) = delete;

  private:
    // Ye Olde Queue.
    std::queue<T*> m_queue;

    bool m_closed;
  };

}

#endif

