#ifndef __MD_LIB_STRINGMESSAGE_H__
#define __MD_LIB_STRINGMESSAGE_H__

#include <ostream>
#include <list>
#include <map>
#include <string>
#include <vector>

#include <MD/MD_sys_types.h>
#include <boost/lexical_cast.hpp>

// MD_StringMessage - associative string-based message class PGW - 3/21/07
// You can store/retrieve any data type that has an iostream (<< and >>)
// representation, similar to stdfds' ConfigFile.
// The interface currently supports only one value per key but could easily
// be expanded to accommodate a list of values. Alternatively, you could just
// just set a field to your own collection of elements, provided the collection
// has an ostream representation.
// Field order is maintained with an internal list of keys, appended as they're
// added.
// Thanks to Anit for the idea and [to|from]_string functionality.

namespace MD {

class MD_StringMessage
{
  public:
  // Static variable to control binary format
  // Versions mismatch while deserializing will fail and return false
  static MD_LIB_API const short SERIALIZATION_VERSION; // defaults to 0

  MD_LIB_API MD_StringMessage();
  MD_LIB_API void clear();

  template<typename T>
  bool get(const std::string &key, T &value) const;
  
  template<typename T>
  bool set(const std::string &key, const T &value);

  // appends to an existing value or sets a new one, delim allows caller
  // to separate the values (e.g. pass in ',')
  template<typename T>
  bool append(const std::string &key, const T &value, char delim = 0);

  // prepends to an existing value or sets a new one, delim allows caller
  // to separate the values (e.g. pass in ',')
  template<typename T>
  bool prepend(const std::string &key, const T &value, char delim = 0);

  MD_LIB_API bool erase (const std::string &key);
  MD_LIB_API bool exists(const std::string &key) const;

  MD_LIB_API size_t size() const;  // size of binary representation
  MD_LIB_API int   count() const;  // number of fields
  MD_LIB_API void  get_keys(std::vector<std::string> &keys) const;

  MD_LIB_API bool serialize(void* ptr) const;
  MD_LIB_API bool deserialize(const void* ptr, size_t size, bool append = false);

  private:
  typedef std::map<std::string, std::string> FieldCollection;
  typedef std::list<std::string> KeyList;

  FieldCollection m_message;
  KeyList         m_keys;
  size_t          m_size;

  friend MD_LIB_API
  std::ostream &operator<<(std::ostream &os, const MD_StringMessage &m);
};

MD_LIB_API std::ostream &operator<<(std::ostream &o, const MD_StringMessage &m);

template<>
MD_LIB_API bool MD_StringMessage::set<std::string>(const std::string &key,
                                        const std::string &val);
template<>
MD_LIB_API bool MD_StringMessage::append<std::string>(const std::string &key,
                                           const std::string &val,
                                           char delim);
template<>
MD_LIB_API bool MD_StringMessage::prepend<std::string>(const std::string &key,
                                            const std::string &val,
                                            char delim);
//
// Templated Function Implementations
//
template <typename T>
inline bool MD_StringMessage::get(const std::string & key, T &value) const
{
  FieldCollection::const_iterator iter = m_message.find(key);
  if(iter == m_message.end()) {
    return false;
  }
  try {
    value = boost::lexical_cast<T>(iter->second);
  }
  catch (const boost::bad_lexical_cast&) {
  }
  return true;
}

template <typename T>
inline bool MD_StringMessage::set(const std::string &key, const T &value)
{ 
  std::string strVal;
  try {
    strVal = boost::lexical_cast<std::string>(value);
  }
  catch (const boost::bad_lexical_cast&) {
  }
  return set<std::string>(key, strVal); 
}

template <typename T>
inline bool MD_StringMessage::append(const std::string &key, const T &value, char delim)
{ 
  std::string strVal;
  try {
    strVal = boost::lexical_cast<std::string>(value);
  }
  catch (const boost::bad_lexical_cast&) {
  }
  return append<std::string>(key, strVal, delim); 
}

template <typename T>
inline bool MD_StringMessage::prepend(const std::string &key, const T &value, char delim)
{ 
  std::string strVal;
  try {
    strVal = boost::lexical_cast<std::string>(value);
  }
  catch (const boost::bad_lexical_cast&) {
  }
  return prepend<std::string>(key, strVal, delim); 
}

inline bool MD_StringMessage::exists(const std::string &key) const
{ return (m_message.find(key) != m_message.end()); }

} // namespace

#endif // __MD_LIB_STRINGMESSAGE_H__
